#!/usr/bin/perl -T

use strict;
use warnings;

#*******************************************************************************
# Data::Filter module/class tests
#
# Author:   Jan Mach, jan.mach@cesnet.cz
# Version:  0.1
# Purpose:
#
#*******************************************************************************

use Data::Dumper;
#use Smart::Comments;

use Test::More tests => 38;

# First test if the module correctly compiles
BEGIN {
        use vars qw($TESTED_CLASS);
        $TESTED_CLASS = 'Data::Filter';

        use_ok('Tester');
        use_ok('Mentat::Message::Factory');
        use_ok('Mentat::Message::Value::IDEA');
        use_ok($TESTED_CLASS);
    }

# Debug output explicit ON switch
#Tester::DEBUG_ON();
#Tester::DEBUG_ON(10);

# Debug output explicit OFF switch
#Tester::DEBUG_OFF();

# User should know, which version is beeing tested (just for sure)
diag("$TESTED_CLASS version: " . $Data::Filter::VERSION) if Tester::D();

# Create test ALERT message
my $message  = Mentat::Message::Factory->from_file('./data/msg_val01.idea');
my $message2 = Mentat::Message::Factory->from_file('./data/msg_val02.idea');

# Configure the test filter configurations
my ($instance, $counter);
my @TESTS = (
        ['./conf/filter-rules.json', 1, 'EXISTS _CESNET', 1],
        [{
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
        }, 1, 'EXISTS _CESNET', 1],
        [{
          'order' => 'deny',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
        }, 0, '((Description EQ "Scan IPMI") AND (Source/IP4 EQ "192.0.2.73"))', 2],
        [{
          'compile' => 'idea',
          'order' => 'deny',
          'deny' => [
                      '? _TEST',
                      'Source/IP4 eq "192.0.2.0/24"',
                    ],
          'allow' => [
                       '? _CESNETS',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.128"',
                     ],
        }, 0, '(Source/IP4 EQ "192.0.2.0/24")', 2],
        [{
          'logop_allow' => 'and',
          'order' => 'allow',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'deny'
        }, 1, '__CHAIN__', 0],
        [{
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       'Source/IP4 == "192.0.2.72"',
                       'Source/IP4 == "192.0.2.73"',
                     ],
        }, 1, '(Source/IP4 EQ "192.0.2.73")', 2],
        [{
          'compile' => 'idea',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       'Source/IP4 == "195.113.144.194"',
                       'CreateTime <= "2015-02-16T02:59:04Z"',
                       'CreateTime >= "2015-02-16T19:59:04Z"',
                       'CreateTime >= "2015-02-16T09:50:04Z"',
                     ],
        }, 1, '(CreateTime GE "2015-02-16 09:50:04Z")', 4],
        [{
          'compile' => 'idea',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       'Source/IP4 in ["195.113.144.192","195.113.144.193",192.0.2.73]',
                     ],
        }, 1, '(Source/IP4 IN ["195.113.144.192","195.113.144.193","192.0.2.73"])', 1],
        [{
          'compile' => 'idea',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       'Category in ["Malware","Test"]',
                     ],
        }, 1, '(Category IN ["Malware","Test"])', 1],
        [{
          'compile' => 'idea',
          'order' => 'deny',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and _CESNET/EventTemplate like "^s\werv"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Alert/@messageid == "1234" and Alert/Target/@ident eq "d1c2b3a4"',
                     ],
        }, 0, '((Description EQ "Scan IPMI") AND (_CESNET/EventTemplate LIKE "(?^:^s\werv)"))', 2],
        [{
          'compile' => 'idea',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and _CESNET/EventTemplate like "^kippo"',
                    ],
          'allow' => [
                       'Source/IP4 like "192.0.\d.[127]3"',
                     ],
        }, 1, '(Source/IP4 LIKE "(?^:192.0.\d.[127]3)")', 1],
        [{
          'compile' => 'idea',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       'Source/IP4 == "192.0.2.72"',
                       'Source/IP4 IN ["192.0.20.70-192.0.20.75","192.0.2.70-192.0.2.75"]',
                       'Source/IP4 == "192.0.2.73"',
                     ],
        }, 1, '(Source/IP4 IN ["192.0.20.70..192.0.20.75","192.0.2.70..192.0.2.75"])', 2],
    );

# Run all above configured tests
$counter = 0;
foreach my $test (@TESTS) {
    $counter++;

    # Now attempt to create a class instance and check it
    $instance = $TESTED_CLASS->new(filter_grammar_file => './conf/filter-grammar.conf',
                                   filter_schema_file => '../../conf/schema-data-filter.json',
                                   filter_config => $test->[0],
                                   dt_resolver => Mentat::Message::Value::IDEA->new());
    isa_ok($instance, $TESTED_CLASS);

    # View the created instance
    diag("Instance dump $counter: " . Dumper($instance)) if Tester::D(2);

    is_deeply([$instance->filter($message)], [$test->[1],$test->[2],$test->[3]], "Filter test $counter");
}

# Configure the failure test filter configurations
my @FAILS = (
    # Check the validation of filter loaded from external file
    ['./conf/filter-rules-inv.json', "Filter configuration file './conf/filter-rules-inv.json' syntax error: ''#/order': invalid enum value 'allowss', allowed values are: 'allow, allow, deny, allow,deny, deny, deny, allow, deny,allow''\n"],
    # Direct validation - extra attribute
    [{
          'compile' => 'idea',
          'orders' => 'deny',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and _CESNET/EventTemplate like "^s\werv"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Alert/@messageid == "1234" and Alert/Target/@ident eq "d1c2b3a4"',
                     ],
    }, "Filter specification syntax error: ''#': invalid property 'orders', no additional properties are allowed'\n"],
    # Direct validation - invalid 'logop_allow' value
    [{
          'logop_allow' => 'andd',
          'order' => 'allow',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'deny'
    }, "Filter specification syntax error: ''#/logop_allow': invalid enum value 'andd', allowed values are: 'and, or''\n"],
    # Direct validation - invalid 'logop_deny' value
    [{
          'logop_deny' => 'andd',
          'order' => 'allow',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'deny'
    }, "Filter specification syntax error: ''#/logop_deny': invalid enum value 'andd', allowed values are: 'and, or''\n"],
    # Direct validation - invalid 'order' value
    [{
          'logop_deny' => 'and',
          'order' => 'alloww',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'deny'
    }, "Filter specification syntax error: ''#/order': invalid enum value 'alloww', allowed values are: 'allow, allow, deny, allow,deny, deny, deny, allow, deny,allow''\n"],
    # Direct validation - invalid 'default_action' value
    [{
          'logop_deny' => 'and',
          'order' => 'allow',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'denyy'
    }, "Filter specification syntax error: ''#/default_action': invalid enum value 'denyy', allowed values are: 'allow, deny''\n"],
    # Direct validation - invalid 'deny' value
    [{
          'logop_deny' => 'and',
          'order' => 'allow',
          'deny' => 'x',
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'deny'
    }, "Filter specification syntax error: ''#/deny': invalid value 'x', expected 'array''\n"],
    # Direct validation - invalid 'allow' value
    [{
          'logop_deny' => 'and',
          'order' => 'allow',
          'deny' => [
                      '? _TEST',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => 'x',
          'default_action' => 'deny'
    }, "Filter specification syntax error: ''#/allow': invalid value 'x', expected 'array''\n"],
    # Direct validation - invalid filter syntax
    [{
          'logop_deny' => 'and',
          'order' => 'allow',
          'deny' => [
                      '? "_TEST"',
                      'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                    ],
          'allow' => [
                       '? _CESNET',
                       'Description == "Scan IPMI" and Source/IP4 eq "192.0.2.73"',
                     ],
          'default_action' => 'deny'
    }, undef], # This test message is too long,
);
# Run all above configured failure tests
$counter = 0;
foreach my $test (@FAILS) {
    $counter++;

    $@ = undef;
    eval {
        $instance = $TESTED_CLASS->new(filter_grammar_file => './conf/filter-grammar.conf',
                                       filter_schema_file => '../../conf/schema-data-filter.json',
                                       filter_config => $test->[0],
                                       dt_resolver => Mentat::Message::Value::IDEA->new());
    };
    # Check match for the explicitly given errors
    if ($test->[1]) {
        is($@, $test->[1], "Filter failure test $counter");
    }
    # Check general failures
    else {
        isnt($@, '', "Filter failure test $counter");
    }
}

#*******************************************************************************

my $fargs = {
            'order' => 'deny',
            'deny' => [
                '(Description IN ["Open NTP Report" , "Scan NTP"])',
            ],
            'default_action' => 'allow',
        };
        
# Now attempt to create a class instance and check it
$instance = $TESTED_CLASS->new(filter_grammar_file => './conf/filter-grammar.conf',
                               filter_schema_file => '../../conf/schema-data-filter.json',
                               filter_config => $fargs,
                               dt_resolver => Mentat::Message::Value::IDEA->new());
isa_ok($instance, $TESTED_CLASS);

# View the created instance
diag("Instance dump X: " . Dumper($instance)) if Tester::D(2);
diag("Instance print X: ".$instance) if Tester::D();
diag("Message dump X: " . Dumper($message2)) if Tester::D();

my $res = [$instance->filter($message2)];
diag("Result dump X: " . Dumper($res)) if Tester::D();

#is_deeply(, [$test->[1],$test->[2],$test->[3]], "Filter test $counter");

#*******************************************************************************
